@@ -1,5 +1,4 @@
 # -*- python-mode -*-
-# -*- coding: UTF-8 -*-
 
 ## Copyright (C) 2012-2013  Daniel Pavel
 ##
@@ -21,14 +20,12 @@
 ## USB product IDs for receivers: 0xC526 - 0xC5xx
 ## Wireless PIDs for hidpp10 devices: 0x2006 - 0x2019
 ## Wireless PIDs for hidpp20 devices: 0x4002 - 0x4097, 0x4101 - 0x4102
-## USB product IDs for hidpp20 devices: 0xC07D - 0xC093, 0xC32B - 0xC344
+## USB product IDs for hidpp20 devices: 0xC07D - 0xC094, 0xC32B - 0xC344
 ## Bluetooth product IDs (for hidpp20 devices): 0xB012 - 0xB0xx, 0xB32A - 0xB3xx
 
 # USB ids of Logitech wireless receivers.
 # Only receivers supporting the HID++ protocol can go in here.
 
-from __future__ import absolute_import, division, print_function, unicode_literals
-
 from .descriptors import DEVICES as _DEVICES
 from .i18n import _
 
@@ -42,12 +39,24 @@ from .i18n import _
 
 _DRIVER = ('hid-generic', 'generic-usb', 'logitech-djreceiver')
 
+_bolt_receiver = lambda product_id: {
+    'vendor_id': 0x046d,
+    'product_id': product_id,
+    'usb_interface': 2,
+    'hid_driver': _DRIVER,  # noqa: F821
+    'name': _('Bolt Receiver'),
+    'receiver_kind': 'bolt',
+    'max_devices': 6,
+    'may_unpair': True
+}
+
 _unifying_receiver = lambda product_id: {
     'vendor_id': 0x046d,
     'product_id': product_id,
     'usb_interface': 2,
     'hid_driver': _DRIVER,  # noqa: F821
-    'name': _('Unifying Receiver')
+    'name': _('Unifying Receiver'),
+    'receiver_kind': 'unifying'
 }
 
 _nano_receiver = lambda product_id: {
@@ -56,6 +65,7 @@ _nano_receiver = lambda product_id: {
     'usb_interface': 1,
     'hid_driver': _DRIVER,  # noqa: F821
     'name': _('Nano Receiver'),
+    'receiver_kind': 'nano',
     'may_unpair': False,
     're_pairs': True
 }
@@ -66,6 +76,7 @@ _nano_receiver_no_unpair = lambda product_id: {
     'usb_interface': 1,
     'hid_driver': _DRIVER,  # noqa: F821
     'name': _('Nano Receiver'),
+    'receiver_kind': 'nano',
     'may_unpair': False,
     'unpair': False,
     're_pairs': True
@@ -77,6 +88,7 @@ _nano_receiver_max2 = lambda product_id: {
     'usb_interface': 1,
     'hid_driver': _DRIVER,  # noqa: F821
     'name': _('Nano Receiver'),
+    'receiver_kind': 'nano',
     'max_devices': 2,
     'may_unpair': False,
     're_pairs': True
@@ -88,6 +100,7 @@ _nano_receiver_maxn = lambda product_id, max: {
     'usb_interface': 1,
     'hid_driver': _DRIVER,  # noqa: F821
     'name': _('Nano Receiver'),
+    'receiver_kind': 'nano',
     'max_devices': max,
     'may_unpair': False,
     're_pairs': True
@@ -98,7 +111,8 @@ _lenovo_receiver = lambda product_id: {
     'product_id': product_id,
     'usb_interface': 1,
     'hid_driver': _DRIVER,  # noqa: F821
-    'name': _('Nano Receiver')
+    'name': _('Nano Receiver'),
+    'receiver_kind': 'nano'
 }
 
 _lightspeed_receiver = lambda product_id: {
@@ -115,67 +129,75 @@ _ex100_receiver = lambda product_id: {
     'usb_interface': 1,
     'hid_driver': _DRIVER,  # noqa: F821
     'name': _('EX100 Receiver 27 Mhz'),
+    'receiver_kind': '27Mhz',
     'max_devices': 4,
     'may_unpair': False,
     're_pairs': True,
     'ex100_27mhz_wpid_fix': True
 }
 
+# Receivers added here should also be listed in
+# share/solaar/io.github.pwr_solaar.solaar.metainfo.xml
+
+# Bolt receivers (marked with the yellow lightning bolt logo)
+BOLT_RECEIVER_C548 = _bolt_receiver(0xc548)
+
 # standard Unifying receivers (marked with the orange Unifying logo)
 UNIFYING_RECEIVER_C52B = _unifying_receiver(0xc52b)
 UNIFYING_RECEIVER_C532 = _unifying_receiver(0xc532)
 
-# Nano receviers that support the Unifying protocol
+# Nano receivers (usually sold with low-end devices)
 NANO_RECEIVER_ADVANCED = _nano_receiver_no_unpair(0xc52f)
-
-# ex100 old style receiver pre-unifyimg protocol
-EX100_27MHZ_RECEIVER_C517 = _ex100_receiver(0xc517)
-
-# Nano receivers that don't support the Unifying protocol
 NANO_RECEIVER_C518 = _nano_receiver(0xc518)
 NANO_RECEIVER_C51A = _nano_receiver(0xc51a)
 NANO_RECEIVER_C51B = _nano_receiver(0xc51b)
 NANO_RECEIVER_C521 = _nano_receiver(0xc521)
 NANO_RECEIVER_C525 = _nano_receiver(0xc525)
 NANO_RECEIVER_C526 = _nano_receiver(0xc526)
-NANO_RECEIVER_C52e = _nano_receiver_no_unpair(0xc52e)
+NANO_RECEIVER_C52E = _nano_receiver_no_unpair(0xc52e)
 NANO_RECEIVER_C531 = _nano_receiver(0xc531)
 NANO_RECEIVER_C534 = _nano_receiver_max2(0xc534)
 NANO_RECEIVER_C537 = _nano_receiver(0xc537)
+NANO_RECEIVER_C542 = _nano_receiver(0xc542)
 NANO_RECEIVER_6042 = _lenovo_receiver(0x6042)
 
-# Lightspeed receivers
+# Lightspeed receivers (usually sold with gaming devices)
 LIGHTSPEED_RECEIVER_C539 = _lightspeed_receiver(0xc539)
-LIGHTSPEED_RECEIVER_C53a = _lightspeed_receiver(0xc53a)
-LIGHTSPEED_RECEIVER_C53f = _lightspeed_receiver(0xc53f)
-LIGHTSPEED_RECEIVER_C53d = _lightspeed_receiver(0xc53d)
-LIGHTSPEED_RECEIVER_C545 = _lightspeed_receiver(0xc545)
+LIGHTSPEED_RECEIVER_C53A = _lightspeed_receiver(0xc53a)
+LIGHTSPEED_RECEIVER_C53D = _lightspeed_receiver(0xc53d)
+LIGHTSPEED_RECEIVER_C53F = _lightspeed_receiver(0xc53f)
 LIGHTSPEED_RECEIVER_C541 = _lightspeed_receiver(0xc541)
+LIGHTSPEED_RECEIVER_C545 = _lightspeed_receiver(0xc545)
 LIGHTSPEED_RECEIVER_C547 = _lightspeed_receiver(0xc547)
 
+# EX100 old style receiver pre-unifying protocol
+EX100_27MHZ_RECEIVER_C517 = _ex100_receiver(0xc517)
+
 ALL = (
+    BOLT_RECEIVER_C548,
     UNIFYING_RECEIVER_C52B,
     UNIFYING_RECEIVER_C532,
     NANO_RECEIVER_ADVANCED,
-    EX100_27MHZ_RECEIVER_C517,
     NANO_RECEIVER_C518,
     NANO_RECEIVER_C51A,
     NANO_RECEIVER_C51B,
     NANO_RECEIVER_C521,
     NANO_RECEIVER_C525,
     NANO_RECEIVER_C526,
-    NANO_RECEIVER_C52e,
+    NANO_RECEIVER_C52E,
     NANO_RECEIVER_C531,
     NANO_RECEIVER_C534,
     NANO_RECEIVER_C537,
+    NANO_RECEIVER_C542,
     NANO_RECEIVER_6042,
     LIGHTSPEED_RECEIVER_C539,
-    LIGHTSPEED_RECEIVER_C53a,
-    LIGHTSPEED_RECEIVER_C53f,
-    LIGHTSPEED_RECEIVER_C53d,
-    LIGHTSPEED_RECEIVER_C545,
+    LIGHTSPEED_RECEIVER_C53A,
+    LIGHTSPEED_RECEIVER_C53D,
+    LIGHTSPEED_RECEIVER_C53F,
     LIGHTSPEED_RECEIVER_C541,
+    LIGHTSPEED_RECEIVER_C545,
     LIGHTSPEED_RECEIVER_C547,
+    EX100_27MHZ_RECEIVER_C517,
 )
 
 _wired_device = lambda product_id, interface: {
@@ -203,10 +225,10 @@ def other_device_check(bus_id, vendor_id, product_id):
     if vendor_id != 0x46d:  # Logitech
         return
     if bus_id == 0x3:  # USB
-        if (product_id >= 0xC07D and product_id <= 0xC093 or product_id >= 0xC32B and product_id <= 0xC344):
+        if (product_id >= 0xC07D and product_id <= 0xC094 or product_id >= 0xC32B and product_id <= 0xC344):
             return _wired_device(product_id, 2)
     elif bus_id == 0x5:  # Bluetooth
-        if (product_id >= 0xB012 and product_id <= 0xB0FF or product_id >= 0xB32A and product_id <= 0xB3FF):
+        if (product_id >= 0xB012 and product_id <= 0xB0FF or product_id >= 0xB317 and product_id <= 0xB3FF):
             return _bt_device(product_id)
 
 
