@@ -1,5 +1,4 @@
 # -*- python-mode -*-
-# -*- coding: UTF-8 -*-
 
 ## Copyright (C) 2012-2013  Daniel Pavel
 ##
@@ -19,30 +18,11 @@
 
 # Some common functions and types.
 
-from __future__ import absolute_import, division, print_function, unicode_literals
-
 from binascii import hexlify as _hexlify
 from collections import namedtuple
 from struct import pack, unpack
 
-try:
-    unicode  # noqa: F821
-    # if Python2, unicode_literals will mess our first (un)pack() argument
-    _pack_str = pack
-    _unpack_str = unpack
-    pack = lambda x, *args: _pack_str(str(x), *args)
-    unpack = lambda x, *args: _unpack_str(str(x), *args)
-
-    is_string = lambda d: isinstance(d, unicode) or isinstance(d, str)  # noqa: F821
-    # no easy way to distinguish between b'' and '' :(
-    # or (isinstance(d, str) \
-    #     and not any((chr(k) in d for k in range(0x00, 0x1F))) \
-    #     and not any((chr(k) in d for k in range(0x80, 0xFF))) \
-    #     )
-except Exception:
-    # this is certainly Python 3
-    # In Py3, unicode and str are equal (the unicode object does not exist)
-    is_string = lambda d: isinstance(d, str)
+is_string = lambda d: isinstance(d, str)
 
 #
 #
@@ -83,13 +63,11 @@ class NamedInt(int):
     def __str__(self):
         return self.name
 
-    __unicode__ = __str__
-
     def __repr__(self):
         return 'NamedInt(%d, %r)' % (int(self), self.name)
 
 
-class NamedInts(object):
+class NamedInts:
     """An ordered set of NamedInt values.
 
     Indexing can be made by int or string, and will return the corresponding
@@ -102,30 +80,33 @@ class NamedInts(object):
     if the value already exists in the set (int or string), ValueError will be
     raised.
     """
-    __slots__ = ('__dict__', '_values', '_indexed', '_fallback')
+    __slots__ = ('__dict__', '_values', '_indexed', '_fallback', '_is_sorted')
 
-    def __init__(self, **kwargs):
+    def __init__(self, dict=None, **kwargs):
         def _readable_name(n):
             if not is_string(n):
-                raise TypeError('expected (unicode) string, got ' + str(type(n)))
+                raise TypeError('expected string, got ' + str(type(n)))
             return n.replace('__', '/').replace('_', ' ')
 
         # print (repr(kwargs))
-        values = {k: NamedInt(v, _readable_name(k)) for (k, v) in kwargs.items()}
+        elements = dict if dict else kwargs
+        values = {k: NamedInt(v, _readable_name(k)) for (k, v) in elements.items()}
         self.__dict__ = values
-        self._values = sorted(list(values.values()))
+        self._is_sorted = False
+        self._values = list(values.values())
+        self._sort_values()
         self._indexed = {int(v): v for v in self._values}
         # assert len(values) == len(self._indexed)
         # "(%d) %r\n=> (%d) %r" % (len(values), values, len(self._indexed), self._indexed)
         self._fallback = None
 
     @classmethod
-    def list(cls, items, name_generator=lambda x: str(x)):
+    def list(cls, items, name_generator=lambda x: str(x)):  # noqa: B008
         values = {name_generator(x): x for x in items}
         return NamedInts(**values)
 
     @classmethod
-    def range(cls, from_value, to_value, name_generator=lambda x: str(x), step=1):
+    def range(cls, from_value, to_value, name_generator=lambda x: str(x), step=1):  # noqa: B008
         values = {name_generator(x): x for x in range(from_value, to_value + 1, step)}
         return NamedInts(**values)
 
@@ -140,14 +121,20 @@ class NamedInts(object):
         if unknown_bits:
             yield 'unknown:%06X' % unknown_bits
 
+    def _sort_values(self):
+        self._values = sorted(self._values)
+        self._is_sorted = True
+
     def __getitem__(self, index):
         if isinstance(index, int):
             if index in self._indexed:
                 return self._indexed[int(index)]
-            if self._fallback and isinstance(index, int):
+            if self._fallback:
                 value = NamedInt(index, self._fallback(index))
                 self._indexed[index] = value
-                self._values = sorted(self._values + [value])
+                self._values.append(value)
+                self._is_sorted = False
+                self._sort_values()
                 return value
 
         elif is_string(index):
@@ -156,21 +143,24 @@ class NamedInts(object):
             return (next((x for x in self._values if str(x) == index), None))
 
         elif isinstance(index, slice):
+            values = self._values if self._is_sorted else sorted(self._values)
+
             if index.start is None and index.stop is None:
-                return self._values[:]
+                return values[:]
 
-            v_start = int(self._values[0]) if index.start is None else int(index.start)
-            v_stop = (self._values[-1] + 1) if index.stop is None else int(index.stop)
+            v_start = int(values[0]) if index.start is None else int(index.start)
+            v_stop = (values[-1] + 1) if index.stop is None else int(index.stop)
 
-            if v_start > v_stop or v_start > self._values[-1] or v_stop <= self._values[0]:
+            if v_start > v_stop or v_start > values[-1] or v_stop <= values[0]:
                 return []
 
-            if v_start <= self._values[0] and v_stop > self._values[-1]:
-                return self._values[:]
+            if v_start <= values[0] and v_stop > values[-1]:
+                return values[:]
 
             start_index = 0
-            stop_index = len(self._values)
-            for i, value in enumerate(self._values):
+            stop_index = len(values)
+
+            for i, value in enumerate(values):
                 if value < v_start:
                     start_index = i + 1
                 elif index.stop is None:
@@ -179,7 +169,7 @@ class NamedInts(object):
                     stop_index = i
                     break
 
-            return self._values[start_index:stop_index]
+            return values[start_index:stop_index]
 
     def __setitem__(self, index, name):
         assert isinstance(index, int), type(index)
@@ -196,19 +186,22 @@ class NamedInts(object):
         if int(value) in self._indexed:
             raise ValueError('%d (%s) already known' % (int(value), value))
 
-        self._values = sorted(self._values + [value])
+        self._values.append(value)
+        self._is_sorted = False
+        self._sort_values()
         self.__dict__[str(value)] = value
         self._indexed[int(value)] = value
 
     def __contains__(self, value):
-        if isinstance(value, int):
+        if isinstance(value, NamedInt):
+            return self[value] == value
+        elif isinstance(value, int):
             return value in self._indexed
         elif is_string(value):
             return value in self.__dict__ or value in self._values
 
     def __iter__(self):
-        for v in self._values:
-            yield v
+        yield from self._values
 
     def __len__(self):
         return len(self._values)
@@ -216,6 +209,18 @@ class NamedInts(object):
     def __repr__(self):
         return 'NamedInts(%s)' % ', '.join(repr(v) for v in self._values)
 
+    def __or__(self, other):
+        return NamedInts(**self.__dict__, **other.__dict__)
+
+
+class UnsortedNamedInts(NamedInts):
+    def _sort_values(self):
+        pass
+
+    def __or__(self, other):
+        c = UnsortedNamedInts if isinstance(other, UnsortedNamedInts) else NamedInts
+        return c(**self.__dict__, **other.__dict__)
+
 
 def strhex(x):
     assert x is not None
@@ -259,11 +264,11 @@ class KwException(Exception):
     They can be later accessed by simple member access.
     """
     def __init__(self, **kwargs):
-        super(KwException, self).__init__(kwargs)
+        super().__init__(kwargs)
 
     def __getattr__(self, k):
         try:
-            return super(KwException, self).__getattr__(k)
+            return super().__getattr__(k)
         except AttributeError:
             return self.args[0][k]
 
