@@ -1,5 +1,4 @@
 # -*- python-mode -*-
-# -*- coding: UTF-8 -*-
 
 ## Copyright (C) 2020 Peter Patel-Schneider
 ##
@@ -17,22 +16,28 @@
 ## with this program; if not, write to the Free Software Foundation, Inc.,
 ## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
+import ctypes as _ctypes
 import os as _os
 import os.path as _path
 import sys as _sys
+import time as _time
 
 from logging import DEBUG as _DEBUG
 from logging import INFO as _INFO
 from logging import getLogger
 from math import sqrt as _sqrt
 
-import _thread
+import evdev
+import keysyms.keysymdef as _keysymdef
 import psutil
 
+from gi.repository import Gdk, GLib
+from solaar.ui.config_panel import change_setting as _change_setting
 from yaml import add_representer as _yaml_add_representer
 from yaml import dump_all as _yaml_dump_all
 from yaml import safe_load_all as _yaml_safe_load_all
 
+from .common import NamedInt
 from .common import unpack as _unpack
 from .hidpp20 import FEATURE as _F
 from .special_keys import CONTROL as _CONTROL
@@ -40,75 +45,177 @@ from .special_keys import CONTROL as _CONTROL
 _log = getLogger(__name__)
 del getLogger
 
-# many of the rule features require X11 so turn rule processing off if X11 is not available
+#
+# See docs/rules.md for documentation
+#
+# Several capabilities of rules depend on aspects of GDK, X11, or XKB
+# As the Solaar GUI uses GTK, Glib and GDK are always available and are obtained from gi.repository
+#
+# Process condition depends on X11 from python-xlib, and is probably not possible at all in Wayland
+# MouseProcess condition depends on X11 from python-xlib, and is probably not possible at all in Wayland
+# Modifiers condition depends only on GDK
+# KeyPress action determines whether a keysym is a currently-down modifier using get_modifier_mapping from python-xlib;
+#   under Wayland no modifier keys are considered down so all modifier keys are pressed, potentially leading to problems
+# KeyPress action translates key names to keysysms using the local file described for GUI keyname determination
+# KeyPress action gets the current keyboard group using XkbGetState from libX11.so using ctypes definitions
+#   under Wayland the keyboard group is None resulting in using the first keyboard group
+# KeyPress action translates keysyms to keycodes using the GDK keymap
+# KeyPress, MouseScroll, and MouseClick actions use XTest (under X11) or uinput.
+# For uinput to work the user must have write access for /dev/uinput.
+# To get this access run  sudo setfacl -m u:${user}:rw /dev/uinput
+#
+# Rule GUI keyname determination uses a local file generated
+#   from http://cgit.freedesktop.org/xorg/proto/x11proto/plain/keysymdef.h
+#   and http://cgit.freedesktop.org/xorg/proto/x11proto/plain/XF86keysym.h
+# because there does not seem to be a non-X11 file for this set of key names
+
+# Setting up is complex because there are several systems that each provide partial facilities:
+# GDK - always available (when running with a window system) but only provides access to keymap
+# X11 - provides access to active process and process with window under mouse and current modifier keys
+# Xtest extension to X11 - provides input simulation, partly works under Wayland
+# Wayland - provides input simulation
+
+XK_KEYS = _keysymdef.keysymdef
+
+# Event codes - can't use Xlib.X codes because Xlib might not be available
+_KEY_RELEASE = 0
+_KEY_PRESS = 1
+_BUTTON_RELEASE = 2
+_BUTTON_PRESS = 3
+
+gdisplay = Gdk.Display.get_default()  # can be None if Solaar is run without a full window system
+gkeymap = Gdk.Keymap.get_for_display(gdisplay) if gdisplay else None
+if _log.isEnabledFor(_INFO):
+    _log.info('GDK Keymap %sset up', '' if gkeymap else 'not ')
+
+wayland = _os.getenv('WAYLAND_DISPLAY')  # is this Wayland?
+if wayland:
+    _log.warn('rules cannot access active process or modifier keys in Wayland')
+
 try:
     import Xlib
-    from Xlib import X
-    from Xlib.display import Display
-    from Xlib.ext import record
-    from Xlib.protocol import rq
-    from Xlib import XK as _XK
-    _XK.load_keysym_group('xf86')
-    XK_KEYS = vars(_XK)
-    disp_prog = Display()
-    x11 = True
-    NET_ACTIVE_WINDOW = disp_prog.intern_atom('_NET_ACTIVE_WINDOW')
-    NET_WM_PID = disp_prog.intern_atom('_NET_WM_PID')
-    WM_CLASS = disp_prog.intern_atom('WM_CLASS')
+    _x11 = None  # X11 might be available
 except Exception:
-    _log.warn('X11 not available - rules will not be activated', exc_info=_sys.exc_info())
-    XK_KEYS = {}
-    x11 = False
+    _x11 = False  # X11 is not available
 
-# determine current key modifiers
-# there must be a better way to do this
+xtest_available = True  # Xtest might be available
+xdisplay = None
+Xkbdisplay = None  # xkb might be available
+modifier_keycodes = []
+XkbUseCoreKbd = 0x100
 
-if x11:
-    display = Display()
-    context = display.record_create_context(
-        0, [record.AllClients], [{
-            'core_requests': (0, 0),
-            'core_replies': (0, 0),
-            'ext_requests': (0, 0, 0, 0),
-            'ext_replies': (0, 0, 0, 0),
-            'delivered_events': (0, 0),
-            'device_events': (X.KeyPress, X.KeyRelease),
-            'errors': (0, 0),
-            'client_started': False,
-            'client_died': False,
-        }]
-    )
-    modifier_keycodes = display.get_modifier_mapping()
-    current_key_modifiers = 0
+
+class XkbDisplay(_ctypes.Structure):
+    """ opaque struct """
+
+
+class XkbStateRec(_ctypes.Structure):
+    _fields_ = [('group', _ctypes.c_ubyte), ('locked_group', _ctypes.c_ubyte), ('base_group', _ctypes.c_ushort),
+                ('latched_group', _ctypes.c_ushort), ('mods', _ctypes.c_ubyte), ('base_mods', _ctypes.c_ubyte),
+                ('latched_mods', _ctypes.c_ubyte), ('locked_mods', _ctypes.c_ubyte), ('compat_state', _ctypes.c_ubyte),
+                ('grab_mods', _ctypes.c_ubyte), ('compat_grab_mods', _ctypes.c_ubyte), ('lookup_mods', _ctypes.c_ubyte),
+                ('compat_lookup_mods', _ctypes.c_ubyte),
+                ('ptr_buttons', _ctypes.c_ushort)]  # something strange is happening here but it is not being used
+
+
+def x11_setup():
+    global _x11, xdisplay, modifier_keycodes, NET_ACTIVE_WINDOW, NET_WM_PID, WM_CLASS, xtest_available
+    if _x11 is not None:
+        return _x11
+    try:
+        from Xlib.display import Display
+        xdisplay = Display()
+        modifier_keycodes = xdisplay.get_modifier_mapping()  # there should be a way to do this in Gdk
+        NET_ACTIVE_WINDOW = xdisplay.intern_atom('_NET_ACTIVE_WINDOW')
+        NET_WM_PID = xdisplay.intern_atom('_NET_WM_PID')
+        WM_CLASS = xdisplay.intern_atom('WM_CLASS')
+        _x11 = True  # X11 available
+        if _log.isEnabledFor(_INFO):
+            _log.info('X11 library loaded and display set up')
+    except Exception:
+        _log.warn('X11 not available - some rule capabilities inoperable: %s', exc_info=_sys.exc_info())
+        _x11 = False
+        xtest_available = False
+    return _x11
+
+
+def xkb_setup():
+    global X11Lib, Xkbdisplay
+    if Xkbdisplay is not None:
+        return Xkbdisplay
+    try:  # set up to get keyboard state using ctypes interface to libx11
+        X11Lib = _ctypes.cdll.LoadLibrary('libX11.so')
+        X11Lib.XOpenDisplay.restype = _ctypes.POINTER(XkbDisplay)
+        X11Lib.XkbGetState.argtypes = [_ctypes.POINTER(XkbDisplay), _ctypes.c_uint, _ctypes.POINTER(XkbStateRec)]
+        Xkbdisplay = X11Lib.XOpenDisplay(None)
+        if _log.isEnabledFor(_INFO):
+            _log.info('XKB display set up')
+    except Exception:
+        _log.warn('XKB display not available - rules cannot access keyboard group: %s', exc_info=_sys.exc_info())
+        Xkbdisplay = False
+    return Xkbdisplay
+
+
+buttons = {
+    'unknown': (None, None),
+    'left': (1, evdev.ecodes.ecodes['BTN_LEFT']),
+    'middle': (2, evdev.ecodes.ecodes['BTN_MIDDLE']),
+    'right': (3, evdev.ecodes.ecodes['BTN_RIGHT']),
+    'scroll_up': (4, evdev.ecodes.ecodes['BTN_4']),
+    'scroll_down': (5, evdev.ecodes.ecodes['BTN_5']),
+    'scroll_left': (6, evdev.ecodes.ecodes['BTN_6']),
+    'scroll_right': (7, evdev.ecodes.ecodes['BTN_7']),
+    'button8': (8, evdev.ecodes.ecodes['BTN_8']),
+    'button9': (9, evdev.ecodes.ecodes['BTN_9']),
+}
+
+# uinput capability for keyboard keys, mouse buttons, and scrolling
+key_events = [c for n, c in evdev.ecodes.ecodes.items() if n.startswith('KEY') and n != 'KEY_CNT']
+for (_, evcode) in buttons.values():
+    if evcode:
+        key_events.append(evcode)
+devicecap = {
+    evdev.ecodes.EV_KEY:
+    key_events,
+    evdev.ecodes.EV_REL:
+    [evdev.ecodes.REL_WHEEL, evdev.ecodes.REL_HWHEEL, evdev.ecodes.REL_WHEEL_HI_RES, evdev.ecodes.REL_HWHEEL_HI_RES]
+}
+udevice = None
+
+
+def setup_uinput():
+    global udevice
+    if udevice is not None:
+        return udevice
+    try:
+        udevice = evdev.uinput.UInput(events=devicecap, name='solaar-keyboard')
+        if _log.isEnabledFor(_INFO):
+            _log.info('uinput device set up')
+    except Exception as e:
+        _log.warn('cannot create uinput device: %s', e)
+
+
+if wayland:  # wayland can't use xtest so may as well set up uinput now
+    setup_uinput()
+
+
+def kbdgroup():
+    if xkb_setup():
+        state = XkbStateRec()
+        X11Lib.XkbGetState(Xkbdisplay, XkbUseCoreKbd, _ctypes.pointer(state))
+        return state.group
+    else:
+        return None
 
 
 def modifier_code(keycode):
-    if keycode == 0:
+    if wayland or not x11_setup() or keycode == 0:
         return None
     for m in range(0, len(modifier_keycodes)):
         if keycode in modifier_keycodes[m]:
             return m
 
 
-def key_press_handler(reply):
-    global current_key_modifiers
-    data = reply.data
-    while len(data):
-        event, data = rq.EventField(None).parse_binary_value(data, display.display, None, None)
-        if event.type == X.KeyPress:
-            mod = modifier_code(event.detail)
-            current_key_modifiers = event.state | 1 << mod if mod is not None else event.state
-        elif event.type == X.KeyRelease:
-            mod = modifier_code(event.detail)
-            current_key_modifiers = event.state & ~(1 << mod) if mod is not None else event.state
-
-
-if x11:
-    _thread.start_new_thread(display.record_enable_context, (context, key_press_handler))
-# display.record_free_context(context)  when should this be done??
-
-# See docs/rules.md for documentation
-
 key_down = None
 key_up = None
 
@@ -144,6 +251,93 @@ def xy_direction(_x, _y):
         return 'noop'
 
 
+def simulate_xtest(code, event):
+    global xtest_available
+    if x11_setup() and xtest_available:
+        try:
+            event = (
+                Xlib.X.KeyPress if event == _KEY_PRESS else Xlib.X.KeyRelease if event == _KEY_RELEASE else
+                Xlib.X.ButtonPress if event == _BUTTON_PRESS else Xlib.X.ButtonRelease if event == _BUTTON_RELEASE else None
+            )
+            Xlib.ext.xtest.fake_input(xdisplay, event, code)
+            xdisplay.sync()
+            if _log.isEnabledFor(_DEBUG):
+                _log.debug('xtest simulated input %s %s %s', xdisplay, event, code)
+            return True
+        except Exception as e:
+            xtest_available = False
+            _log.warn('xtest fake input failed: %s', e)
+
+
+def simulate_uinput(what, code, arg):
+    global udevice
+    if setup_uinput():
+        try:
+            udevice.write(what, code, arg)
+            udevice.syn()
+            if _log.isEnabledFor(_DEBUG):
+                _log.debug('uinput simulated input %s %s %s', what, code, arg)
+            return True
+        except Exception as e:
+            udevice = None
+            _log.warn('uinput write failed: %s', e)
+
+
+def simulate_key(code, event):  # X11 keycode but Solaar event code
+    if not wayland and simulate_xtest(code, event):
+        return True
+    if simulate_uinput(evdev.ecodes.EV_KEY, code - 8, event):
+        return True
+    _log.warn('no way to simulate key input')
+
+
+def click_xtest(button, count):
+    for _ in range(count):
+        if not simulate_xtest(button[0], _BUTTON_PRESS):
+            return False
+        if not simulate_xtest(button[0], _BUTTON_RELEASE):
+            return False
+    return True
+
+
+def click_uinput(button, count):
+    for _ in range(count):
+        if not simulate_uinput(evdev.ecodes.EV_KEY, button[1], 1):
+            return False
+        if not simulate_uinput(evdev.ecodes.EV_KEY, button[1], 0):
+            return False
+    return True
+
+
+def click(button, count):
+    if not wayland and click_xtest(button, count):
+        return True
+    if click_uinput(button, count):
+        return True
+    _log.warn('no way to simulate mouse click')
+    return False
+
+
+def simulate_scroll(dx, dy):
+    if not wayland and xtest_available:
+        success = True
+        if dx:
+            success = click_xtest(buttons['scroll_right' if dx > 0 else 'scroll_left'], count=abs(dx))
+        if dy and success:
+            success = click_xtest(buttons['scroll_up' if dy > 0 else 'scroll_down'], count=abs(dy))
+        if success:
+            return True
+    if setup_uinput():
+        success = True
+        if dx:
+            success = simulate_uinput(evdev.ecodes.EV_REL, evdev.ecodes.REL_HWHEEL, dx)
+        if dy and success:
+            success = simulate_uinput(evdev.ecodes.EV_REL, evdev.ecodes.REL_WHEEL, dy)
+        if success:
+            return True
+    _log.warn('no way to simulate scrolling')
+
+
 TESTS = {
     'crown_right': lambda f, r, d: f == _F.CROWN and r == 0 and d[1] < 128 and d[1],
     'crown_left': lambda f, r, d: f == _F.CROWN and r == 0 and d[1] >= 128 and 256 - d[1],
@@ -173,11 +367,8 @@ MOUSE_GESTURE_TESTS = {
 
 COMPONENTS = {}
 
-if x11:
-    displayt = Display()
 
-
-class RuleComponent(object):
+class RuleComponent:
     def compile(self, c):
         if isinstance(c, RuleComponent):
             return c
@@ -284,34 +475,42 @@ class And(Condition):
 
 
 def x11_focus_prog():
+    if not x11_setup():
+        return None
     pid = wm_class = None
-    window = disp_prog.get_input_focus().focus
+    window = xdisplay.get_input_focus().focus
     while window:
         pid = window.get_full_property(NET_WM_PID, 0)
         wm_class = window.get_wm_class()
-        if wm_class:
+        if wm_class and pid:
             break
         window = window.query_tree().parent
-    name = psutil.Process(pid.value[0]).name() if pid else None
-    return (wm_class[0], wm_class[1], name) if wm_class else (name)
+    try:
+        name = psutil.Process(pid.value[0]).name() if pid else ''
+    except Exception:
+        name = ''
+    return (wm_class[0], wm_class[1], name) if wm_class else (name, )
 
 
 def x11_pointer_prog():
+    if not x11_setup():
+        return None
     pid = wm_class = None
-    window = disp_prog.screen().root.query_pointer().child
-    for window in reversed(window.query_tree().children):
-        pid = window.get_full_property(NET_WM_PID, 0)
-        wm_class = window.get_wm_class()
+    window = xdisplay.screen().root.query_pointer().child
+    for child in reversed(window.query_tree().children):
+        pid = child.get_full_property(NET_WM_PID, 0)
+        wm_class = child.get_wm_class()
         if wm_class:
             break
-        window = window.query_tree().parent
-    name = psutil.Process(pid.value[0]).name() if pid else None
-    return (wm_class[0], wm_class[1], name) if wm_class else (name)
+    name = psutil.Process(pid.value[0]).name() if pid else ''
+    return (wm_class[0], wm_class[1], name) if wm_class else (name, )
 
 
 class Process(Condition):
     def __init__(self, process):
         self.process = process
+        if wayland or not x11_setup():
+            _log.warn('rules can only access active process in X11 - %s', self)
         if not isinstance(process, str):
             _log.warn('rule Process argument not a string: %s', process)
             self.process = str(process)
@@ -322,7 +521,8 @@ class Process(Condition):
     def evaluate(self, feature, notification, device, status, last_result):
         if not isinstance(self.process, str):
             return False
-        result = any(bool(s and s.startswith(self.process)) for s in x11_focus_prog())
+        focus = x11_focus_prog()
+        result = any(bool(s and s.startswith(self.process)) for s in focus) if focus else None
         return result
 
     def data(self):
@@ -332,6 +532,8 @@ class Process(Condition):
 class MouseProcess(Condition):
     def __init__(self, process):
         self.process = process
+        if wayland or not x11_setup():
+            _log.warn('rules cannot access active mouse process in X11 - %s', self)
         if not isinstance(process, str):
             _log.warn('rule MouseProcess argument not a string: %s', process)
             self.process = str(process)
@@ -342,7 +544,8 @@ class MouseProcess(Condition):
     def evaluate(self, feature, notification, device, status, last_result):
         if not isinstance(self.process, str):
             return False
-        result = any(bool(s and s.startswith(self.process)) for s in x11_pointer_prog())
+        pointer_focus = x11_pointer_prog()
+        result = any(bool(s and s.startswith(self.process)) for s in pointer_focus) if pointer_focus else None
         return result
 
     def data(self):
@@ -383,7 +586,49 @@ class Report(Condition):
         return {'Report': self.report}
 
 
-MODIFIERS = {'Shift': 0x01, 'Control': 0x04, 'Alt': 0x08, 'Super': 0x40}
+# Setting(device, setting, [key], value...)
+class Setting(Condition):
+    def __init__(self, args):
+        if not (isinstance(args, list) and len(args) > 2):
+            _log.warn('rule Setting argument not list with minimum length 3: %s', args)
+            self.args = []
+        else:
+            self.args = args
+
+    def __str__(self):
+        return 'Setting: ' + ' '.join([str(a) for a in self.args])
+
+    def evaluate(self, report, notification, device, status, last_result):
+        import solaar.ui.window as _window
+        if len(self.args) < 3:
+            return None
+        dev = _window.find_device(self.args[0]) if self.args[0] is not None else device
+        if dev is None:
+            _log.warn('Setting condition: device %s is not known', self.args[0])
+            return False
+        setting = next((s for s in dev.settings if s.name == self.args[1]), None)
+        if setting is None:
+            _log.warn('Setting condition: setting %s is not the name of a setting for %s', self.args[1], dev.name)
+            return None
+        # should the value argument be checked to be sure it is acceptable?? needs to be careful about boolean toggle
+        # TODO add compare  methods for more validators
+        try:
+            result = setting.compare(self.args[2:], setting.read())
+        except Exception as e:
+            _log.warn('Setting condition: error when checking setting %s: %s', self.args, e)
+            result = False
+        return result
+
+    def data(self):
+        return {'Setting': self.args[:]}
+
+
+MODIFIERS = {
+    'Shift': int(Gdk.ModifierType.SHIFT_MASK),
+    'Control': int(Gdk.ModifierType.CONTROL_MASK),
+    'Alt': int(Gdk.ModifierType.MOD1_MASK),
+    'Super': int(Gdk.ModifierType.MOD4_MASK)
+}
 MODIFIER_MASK = MODIFIERS['Shift'] + MODIFIERS['Control'] + MODIFIERS['Alt'] + MODIFIERS['Super']
 
 
@@ -403,7 +648,12 @@ class Modifiers(Condition):
         return 'Modifiers: ' + str(self.desired)
 
     def evaluate(self, feature, notification, device, status, last_result):
-        return self.desired == (current_key_modifiers & MODIFIER_MASK)
+        if gkeymap:
+            current = gkeymap.get_modifier_state()  # get the current keyboard modifier
+            return self.desired == (current & MODIFIER_MASK)
+        else:
+            _log.warn('no keymap so cannot determine modifier keys')
+            return False
 
     def data(self):
         return {'Modifiers': [str(m) for m in self.modifiers]}
@@ -481,11 +731,10 @@ class Test(Condition):
             else:
                 _log.warn('rule Test string argument not name of a test: %s', test)
                 self.function = TESTS['False']
-        elif (
-            isinstance(test, list) and 2 < len(test) <= 4 and all(isinstance(t, int) for t in test) and test[0] >= 0
-            and test[0] <= 16 and test[1] >= 0 and test[1] <= 16 and test[0] < test[1]
-        ):
-            self.function = bit_test(*test) if len(test) == 3 else range_test(*test)
+        elif isinstance(test, list) and all(isinstance(t, int) for t in test):
+            _log.warn('Test rules consisting of numbers are deprecated, converting to a TestBytes condition')
+            self.__class__ = TestBytes
+            self.__init__(test)
         else:
             _log.warn('rule Test argument not valid %s', test)
 
@@ -499,6 +748,27 @@ class Test(Condition):
         return {'Test': str(self.test)}
 
 
+class TestBytes(Condition):
+    def __init__(self, test):
+        self.test = test
+        if (
+            isinstance(test, list) and 2 < len(test) <= 4 and all(isinstance(t, int) for t in test) and test[0] >= 0
+            and test[0] <= 16 and test[1] >= 0 and test[1] <= 16 and test[0] < test[1]
+        ):
+            self.function = bit_test(*test) if len(test) == 3 else range_test(*test)
+        else:
+            _log.warn('rule TestBytes argument not valid %s', test)
+
+    def __str__(self):
+        return 'TestBytes: ' + str(self.test)
+
+    def evaluate(self, feature, notification, device, status, last_result):
+        return self.function(feature, notification.address, notification.data)
+
+    def data(self):
+        return {'TestBytes': self.test[:]}
+
+
 class MouseGesture(Condition):
     MOVEMENTS = [
         'Mouse Up', 'Mouse Down', 'Mouse Left', 'Mouse Right', 'Mouse Up-left', 'Mouse Up-right', 'Mouse Down-left',
@@ -557,41 +827,57 @@ class KeyPress(Action):
     def __init__(self, keys):
         if isinstance(keys, str):
             keys = [keys]
-        self.key_symbols = keys
-        key_from_string = lambda s: displayt.keysym_to_keycode(Xlib.XK.string_to_keysym(s))
-        self.keys = [isinstance(k, str) and key_from_string(k) for k in keys]
-        if not all(self.keys):
-            _log.warn('rule KeyPress argument not sequence of key names %s', keys)
-            self.keys = []
+        self.key_names = keys
+        self.key_symbols = [XK_KEYS.get(k, None) for k in keys]
+        if not all(self.key_symbols):
+            _log.warn('rule KeyPress not sequence of key names %s', keys)
+            self.key_symbols = []
+
+    def keysym_to_keycode(self, keysym, modifiers):  # maybe should take shift into account
+        group = kbdgroup() or 0
+        keycodes = gkeymap.get_entries_for_keyval(keysym)
+        if len(keycodes.keys) == 1:
+            k = keycodes.keys[0]
+            return k.keycode
+        else:
+            for k in keycodes.keys:
+                if group is None or group == k.group:
+                    return k.keycode
+            _log.warn('rule KeyPress key symbol not currently available %s', self)
 
     def __str__(self):
-        return 'KeyPress: ' + ' '.join(self.key_symbols)
+        return 'KeyPress: ' + ' '.join(self.key_names)
 
-    def needed(self, k, current_key_modifiers):
+    def needed(self, k, modifiers):
         code = modifier_code(k)
-        return not (code and current_key_modifiers & (1 << code))
+        return not (code is not None and modifiers & (1 << code))
 
-    def keyDown(self, keys, modifiers):
-        for k in keys:
-            if self.needed(k, modifiers):
-                Xlib.ext.xtest.fake_input(displayt, X.KeyPress, k)
+    def keyDown(self, keysyms, modifiers):
+        for k in keysyms:
+            keycode = self.keysym_to_keycode(k, modifiers)
+            if keycode and self.needed(keycode, modifiers):
+                simulate_key(keycode, _KEY_PRESS)
 
-    def keyUp(self, keys, modifiers):
-        for k in keys:
-            if self.needed(k, modifiers):
-                Xlib.ext.xtest.fake_input(displayt, X.KeyRelease, k)
+    def keyUp(self, keysyms, modifiers):
+        for k in keysyms:
+            keycode = self.keysym_to_keycode(k, modifiers)
+            if keycode and self.needed(keycode, modifiers):
+                simulate_key(keycode, _KEY_RELEASE)
 
     def evaluate(self, feature, notification, device, status, last_result):
-        current = current_key_modifiers
-        if _log.isEnabledFor(_INFO):
-            _log.info('KeyPress action: %s, modifiers %s %s', self.key_symbols, current, [hex(k) for k in self.keys])
-        self.keyDown(self.keys, current)
-        self.keyUp(reversed(self.keys), current)
-        displayt.sync()
+        if gkeymap:
+            current = gkeymap.get_modifier_state()
+            if _log.isEnabledFor(_INFO):
+                _log.info('KeyPress action: %s, modifiers %s %s', self.key_names, current, [hex(k) for k in self.key_symbols])
+            self.keyDown(self.key_symbols, current)
+            self.keyUp(reversed(self.key_symbols), current)
+            _time.sleep(0.01)
+        else:
+            _log.warn('no keymap so cannot determine which keycode to send')
         return None
 
     def data(self):
-        return {'KeyPress': [str(k) for k in self.key_symbols]}
+        return {'KeyPress': [str(k) for k in self.key_names]}
 
 
 # KeyDown is dangerous as the key can auto-repeat and make your system unusable
@@ -602,25 +888,6 @@ class KeyPress(Action):
 #    def evaluate(self, feature, notification, device, status, last_result):
 #        super().keyUp(self.keys, current_key_modifiers)
 
-buttons = {
-    'unknown': None,
-    'left': 1,
-    'middle': 2,
-    'right': 3,
-    'scroll_up': 4,
-    'scroll_down': 5,
-    'scroll_left': 6,
-    'scroll_right': 7
-}
-for i in range(8, 31):
-    buttons['button%d' % i] = i
-
-
-def click(button, count):
-    for _ in range(count):
-        Xlib.ext.xtest.fake_input(displayt, Xlib.X.ButtonPress, button)
-        Xlib.ext.xtest.fake_input(displayt, Xlib.X.ButtonRelease, button)
-
 
 class MouseScroll(Action):
     def __init__(self, amounts):
@@ -644,11 +911,8 @@ class MouseScroll(Action):
         if _log.isEnabledFor(_INFO):
             _log.info('MouseScroll action: %s %s %s', self.amounts, last_result, amounts)
         dx, dy = amounts
-        if dx:
-            click(button=buttons['scroll_right'] if dx > 0 else buttons['scroll_left'], count=abs(dx))
-        if dy:
-            click(button=buttons['scroll_up'] if dy > 0 else buttons['scroll_down'], count=abs(dy))
-        displayt.sync()
+        simulate_scroll(dx, dy)
+        _time.sleep(0.01)
         return None
 
     def data(self):
@@ -680,13 +944,51 @@ class MouseClick(Action):
             _log.info('MouseClick action: %d %s' % (self.count, self.button))
         if self.button and self.count:
             click(buttons[self.button], self.count)
-        displayt.sync()
+        _time.sleep(0.01)
         return None
 
     def data(self):
         return {'MouseClick': [self.button, self.count]}
 
 
+class Set(Action):
+    def __init__(self, args):
+        if not (isinstance(args, list) and len(args) > 2):
+            _log.warn('rule Set argument not list with minimum length 3: %s', args)
+            self.args = []
+        else:
+            self.args = args
+
+    def __str__(self):
+        return 'Set: ' + ' '.join([str(a) for a in self.args])
+
+    def evaluate(self, feature, notification, device, status, last_result):
+        import solaar.ui.window as _window
+        # importing here to avoid circular imports
+
+        if len(self.args) < 3:
+            return None
+        if _log.isEnabledFor(_INFO):
+            _log.info('Set action: %s', self.args)
+        dev = _window.find_device(self.args[0]) if self.args[0] is not None else device
+        if dev is None:
+            _log.error('Set action: device %s is not known', self.args[0])
+            return None
+        setting = next((s for s in dev.settings if s.name == self.args[1]), None)
+        if setting is None:
+            _log.error('Set action: setting %s is not the name of a setting for %s', self.args[1], dev.name)
+            return None
+        args = setting.acceptable(self.args[2:], setting.read())
+        if args is None:
+            _log.error('Set Action: invalid args %s for setting %s of %s', self.args[2:], self.args[1], self.args[0])
+            return None
+        _change_setting(dev, setting, args)
+        return None
+
+    def data(self):
+        return {'Set': self.args[:]}
+
+
 class Execute(Action):
     def __init__(self, args):
         if isinstance(args, str):
@@ -720,56 +1022,53 @@ COMPONENTS = {
     'MouseProcess': MouseProcess,
     'Feature': Feature,
     'Report': Report,
+    'Setting': Setting,
     'Modifiers': Modifiers,
     'Key': Key,
     'Test': Test,
+    'TestBytes': TestBytes,
     'MouseGesture': MouseGesture,
     'KeyPress': KeyPress,
     'MouseScroll': MouseScroll,
     'MouseClick': MouseClick,
+    'Set': Set,
     'Execute': Execute,
 }
 
 built_in_rules = Rule([])
-if x11:
+if True:
     built_in_rules = Rule([
         {'Rule': [  # Implement problematic keys for Craft and MX Master
-            {'Rule': [{'Key': 'Brightness Down'}, {'KeyPress': 'XF86_MonBrightnessDown'}]},
-            {'Rule': [{'Key': 'Brightness Up'}, {'KeyPress': 'XF86_MonBrightnessUp'}]},
+            {'Rule': [{'Key': ['Brightness Down', 'pressed']}, {'KeyPress': 'XF86_MonBrightnessDown'}]},
+            {'Rule': [{'Key': ['Brightness Up', 'pressed']}, {'KeyPress': 'XF86_MonBrightnessUp'}]},
         ]},
-        {'Rule': [  # In firefox, crown emits keys that move up and down if not pressed, rotate through tabs otherwise
-            {'Process': 'firefox'},
-            {'Rule': [{'Test': 'crown_pressed'}, {'Test': 'crown_right_ratchet'}, {'KeyPress': ['Control_R', 'Tab']}]},
-            {'Rule': [{'Test': 'crown_pressed'},
-                      {'Test': 'crown_left_ratchet'},
-                      {'KeyPress': ['Control_R', 'Shift_R', 'Tab']}]},
-            {'Rule': [{'Test': 'crown_right_ratchet'}, {'KeyPress': 'Down'}]},
-            {'Rule': [{'Test': 'crown_left_ratchet'}, {'KeyPress': 'Up'}]},
-        ]},
-        {'Rule': [  # Otherwise, crown movements emit keys that modify volume if not pressed, move between tracks otherwise
-            {'Feature': 'CROWN'}, {'Report': 0x0},
-            {'Rule': [{'Test': 'crown_pressed'}, {'Test': 'crown_right_ratchet'}, {'KeyPress': 'XF86_AudioNext'}]},
-            {'Rule': [{'Test': 'crown_pressed'}, {'Test': 'crown_left_ratchet'}, {'KeyPress': 'XF86_AudioPrev'}]},
-            {'Rule': [{'Test': 'crown_right_ratchet'}, {'KeyPress': 'XF86_AudioRaiseVolume'}]},
-            {'Rule': [{'Test': 'crown_left_ratchet'}, {'KeyPress': 'XF86_AudioLowerVolume'}]}
-        ]},
-        {'Rule': [  # Thumb wheel does horizontal movement, doubled if control key not pressed
-            {'Feature': 'THUMB WHEEL'},  # with control modifier on mouse scrolling sometimes does something different!
-            {'Rule': [{'Modifiers': 'Control'}, {'Test': 'thumb_wheel_up'}, {'MouseScroll': [-1, 0]}]},
-            {'Rule': [{'Modifiers': 'Control'}, {'Test': 'thumb_wheel_down'}, {'MouseScroll': [-1, 0]}]},
-            {'Rule': [{'Or': [{'Test': 'thumb_wheel_up'}, {'Test': 'thumb_wheel_down'}]}, {'MouseScroll': [-2, 0]}]}
-        ]}
+        # {'Rule': [  # In firefox, crown emits keys that move up and down if not pressed, rotate through tabs otherwise
+        #     {'Process': 'firefox'},
+        #     {'Rule': [{'Test': 'crown_pressed'}, {'Test': 'crown_right_ratchet'}, {'KeyPress': ['Control_R', 'Tab']}]},
+        #     {'Rule': [{'Test': 'crown_pressed'},
+        #               {'Test': 'crown_left_ratchet'},
+        #               {'KeyPress': ['Control_R', 'Shift_R', 'Tab']}]},
+        #     {'Rule': [{'Test': 'crown_right_ratchet'}, {'KeyPress': 'Down'}]},
+        #     {'Rule': [{'Test': 'crown_left_ratchet'}, {'KeyPress': 'Up'}]},
+        # ]},
+        # {'Rule': [  # Otherwise, crown movements emit keys that modify volume if not pressed, move between tracks otherwise
+        #     {'Feature': 'CROWN'}, {'Report': 0x0},
+        #     {'Rule': [{'Test': 'crown_pressed'}, {'Test': 'crown_right_ratchet'}, {'KeyPress': 'XF86_AudioNext'}]},
+        #     {'Rule': [{'Test': 'crown_pressed'}, {'Test': 'crown_left_ratchet'}, {'KeyPress': 'XF86_AudioPrev'}]},
+        #     {'Rule': [{'Test': 'crown_right_ratchet'}, {'KeyPress': 'XF86_AudioRaiseVolume'}]},
+        #     {'Rule': [{'Test': 'crown_left_ratchet'}, {'KeyPress': 'XF86_AudioLowerVolume'}]}
+        # ]},
     ])
 
 keys_down = []
 g_keys_down = [0, 0, 0, 0]
+m_keys_down = 0
+mr_key_down = False
 
 
 # process a notification
 def process_notification(device, status, notification, feature):
-    if not x11:
-        return
-    global keys_down, g_keys_down, key_down, key_up
+    global keys_down, g_keys_down, m_keys_down, mr_key_down, key_down, key_up
     key_down, key_up = None, None
     # need to keep track of keys that are down to find a new key down
     if feature == _F.REPROG_CONTROLS_V4 and notification.address == 0x00:
@@ -793,7 +1092,24 @@ def process_notification(device, status, notification, feature):
                 if old_byte & (0x01 << (i - 1)) and not new_byte & (0x01 << (i - 1)):
                     key_up = _CONTROL['G' + str(i + 8 * byte_idx)]
         g_keys_down = new_g_keys_down
-    rules.evaluate(feature, notification, device, status, True)
+    # and also M keys down
+    elif feature == _F.MKEYS and notification.address == 0x00:
+        new_m_keys_down = _unpack('!1B', notification.data[:1])[0]
+        for i in range(1, 9):
+            if new_m_keys_down & (0x01 << (i - 1)) and not m_keys_down & (0x01 << (i - 1)):
+                key_down = _CONTROL['M' + str(i)]
+            if m_keys_down & (0x01 << (i - 1)) and not new_m_keys_down & (0x01 << (i - 1)):
+                key_up = _CONTROL['M' + str(i)]
+        m_keys_down = new_m_keys_down
+    # and also MR key
+    elif feature == _F.MR and notification.address == 0x00:
+        new_mr_key_down = _unpack('!1B', notification.data[:1])[0]
+        if not mr_key_down and new_mr_key_down:
+            key_down = _CONTROL['MR']
+        if mr_key_down and not new_mr_key_down:
+            key_up = _CONTROL['MR']
+        mr_key_down = new_mr_key_down
+    GLib.idle_add(rules.evaluate, feature, notification, device, status, True)
 
 
 _XDG_CONFIG_HOME = _os.environ.get('XDG_CONFIG_HOME') or _path.expanduser(_path.join('~', '.config'))
@@ -808,7 +1124,7 @@ def _save_config_rule_file(file_name=_file_path):
         pass
 
     def blockseq_rep(dumper, data):
-        return dumper.represent_sequence(u'tag:yaml.org,2002:seq', data, flow_style=True)
+        return dumper.represent_sequence('tag:yaml.org,2002:seq', data, flow_style=True)
 
     _yaml_add_representer(inline_list, blockseq_rep)
 
@@ -822,6 +1138,8 @@ def _save_config_rule_file(file_name=_file_path):
             return [convert(c) for c in elem]
         if isinstance(elem, dict):
             return {k: convert(v) for k, v in elem.items()}
+        if isinstance(elem, NamedInt):
+            return int(elem)
         return elem
 
     # YAML format settings
@@ -833,7 +1151,7 @@ def _save_config_rule_file(file_name=_file_path):
         # 'version': (1, 3),  # it would be printed for every rule
     }
     # Save only user-defined rules
-    rules_to_save = sum([r.data()['Rule'] for r in rules.components if r.source == file_name], [])
+    rules_to_save = sum((r.data()['Rule'] for r in rules.components if r.source == file_name), [])
     if rules_to_save:
         if _log.isEnabledFor(_INFO):
             _log.info('saving %d rule(s) to %s', len(rules_to_save), file_name)
@@ -852,7 +1170,7 @@ def _load_config_rule_file():
     loaded_rules = []
     if _path.isfile(_file_path):
         try:
-            with open(_file_path, 'r') as config_file:
+            with open(_file_path) as config_file:
                 loaded_rules = []
                 for loaded_rule in _yaml_safe_load_all(config_file):
                     rule = Rule(loaded_rule, source=_file_path)
@@ -866,5 +1184,4 @@ def _load_config_rule_file():
     rules = Rule([Rule(loaded_rules, source=_file_path), built_in_rules])
 
 
-if x11:
-    _load_config_rule_file()
+_load_config_rule_file()
